import functools
import inspect
from typing import Callable, List, Any, Dict


async def execute_middlewares(func: Callable, routine_func: Callable, middlewares: List, *args: Any) -> Any:
    if middlewares:
        middleware_context = {}  # type: Dict

        async def middleware_bubble(idx: int = 0, *ma: Any, **mkw: Any) -> Any:
            @functools.wraps(func)
            async def _func(*a: Any, **kw: Any) -> Any:
                return await middleware_bubble(idx + 1, *a, **kw)

            if middlewares and len(middlewares) <= idx + 1:
                _func = routine_func

            middleware = middlewares[idx]  # type: Callable

            arg_len = len(inspect.getfullargspec(middleware).args) - (len(inspect.getfullargspec(middleware).defaults) if inspect.getfullargspec(middleware).defaults else 0)
            middleware_arguments = [_func, *args, middleware_context][0:arg_len]

            return await middleware(*middleware_arguments, *ma, **mkw)

        return_value = await middleware_bubble()
    else:
        return_value = await routine_func()

    return return_value



import types
import functools
from typing import Any, Callable, Optional, Dict, List  # noqa

FUNCTION_ATTRIBUTE = 'TOMODACHI_INVOKER'
START_ATTRIBUTE = 'TOMODACHI_INVOKER_START'


class Invoker(object):
    context = {}  # type: Dict

    @classmethod
    def decorator(cls, cls_func: Callable) -> Callable:
        def _wrapper(*args: Any, **kwargs: Any) -> Callable:
            def wrapper(func: Callable) -> Callable:
                @functools.wraps(func)
                async def _decorator(obj: Any, *a: Any, **kw: Any) -> Any:
                    if not getattr(_decorator, START_ATTRIBUTE, None):
                        return await func(obj, *a, **kw)

                    setattr(_decorator, START_ATTRIBUTE, False)
                    if not cls.context.get(obj, None):
                        cls.context[obj] = {i: getattr(obj, i) for i in dir(obj) if not callable(i) and not i.startswith("__") and not isinstance(getattr(obj, i), types.MethodType)}
                    context = cls.context[obj]
                    obj.context = context
                    start_func = await cls_func(cls, obj, context, func, *args, **kwargs)
                    return start_func

                setattr(_decorator, FUNCTION_ATTRIBUTE, True)
                return _decorator

            if not kwargs and len(args) == 1 and callable(args[0]):
                func = args[0]
                args = ()
                return wrapper(func)
            else:
                return wrapper
        return _wrapper



import os
import sys
import re
import logging
import importlib
import importlib.util
from types import ModuleType
from typing import Any  # noqa


class ServicePackageError(ImportError):
    pass


try:
    if ModuleNotFoundError:
        pass
except Exception:
    class ModuleNotFoundError(ImportError):
        pass


class ServiceImporter(object):
    @classmethod
    def import_service_file(cls, file_name: str) -> ModuleType:
        cwd = os.getcwd()
        file_path = '{}/{}.py'.format(os.path.realpath(cwd), file_name)
        if file_path.endswith('.py.py'):
            file_path = file_path[:-3]
        try:
            sys.path.insert(0, cwd)
            sys.path.insert(0, os.path.dirname(os.path.dirname(file_path)))
            if file_path.endswith('.py') and not os.path.isfile(file_path):
                raise OSError('No such service file')
            elif not file_path.endswith('.py') and not os.path.isfile('{}.py'.format(file_path)):
                raise OSError('No such service file')
            try:
                spec = importlib.util.find_spec('.{}'.format(file_path.rsplit('/', 1)[1])[:-3], package=os.path.dirname(file_path).rsplit('/', 1)[1])  # type: Any
                if not spec:
                    # package name already taken, imported service not part of a package
                    spec = importlib.util.spec_from_file_location(file_name, file_path)
            except AttributeError as e:
                # package name already taken, imported service not part of a package
                try:
                    spec = importlib.util.spec_from_file_location(file_name, file_path)
                except Exception:
                    raise e
            except ModuleNotFoundError as e:  # noqa
                file_path_package_name = file_path[:-3] if file_path.endswith('.py') else file_path
                if str(e) == "__path__ attribute not found on '{}' while trying to find '{}'".format(file_path_package_name.rsplit('/', 2)[1], '.'.join(file_path_package_name.rsplit('/', 2)[1:])):
                    logging.getLogger('import').warning('Invalid service package/parent name, may conflict with Python internals: "{}" - change parent folder name'.format(file_path.rsplit('/', 2)[1]))
                    raise ServicePackageError from e
                if str(e) == "__path__ attribute not found on '{}'while trying to find '{}'".format(file_path_package_name.rsplit('/', 2)[1], '.'.join(file_path_package_name.rsplit('/', 2)[1:])):
                    logging.getLogger('import').warning('Invalid service package/parent name, may conflict with Python internals: "{}" - change parent folder name'.format(file_path.rsplit('/', 2)[1]))
                    raise ServicePackageError from e
                raise e
            if not spec:
                raise OSError
            service_import = importlib.util.module_from_spec(spec)
            try:
                importlib.reload(service_import)
                service_import = importlib.util.module_from_spec(spec)
            except ImportError:
                pass
            if not spec.loader:
                raise OSError
            try:
                if service_import:
                    service_import_name = service_import.__name__[:-3] if service_import.__name__.endswith('.py') else service_import.__name__
                else:
                    service_import_name = ''
                spec.loader.exec_module(service_import)
            except ImportError as e:
                if service_import_name and str(e) == "No module named '{}'".format(service_import_name):
                    logging.getLogger('import').warning('Invalid service package/parent name, may conflict with Python internals: "{}" - change parent folder name'.format(file_path.rsplit('/', 2)[1]))
                    raise ServicePackageError from e
                if str(e) == 'attempted relative import with no known parent package':
                    logging.getLogger('import').warning('Invalid service package/parent name, may conflict with Python internals: "{}" - change parent folder name'.format(file_path.rsplit('/', 2)[1]))
                    raise ServicePackageError from e
                raise e
            except SystemError as e:
                if service_import_name and str(e) == "Parent module '{}' not loaded, cannot perform relative import".format(service_import_name):
                    logging.getLogger('import').warning('Invalid service package/parent name, may conflict with Python internals: "{}" - change parent folder name'.format(file_path.rsplit('/', 2)[1]))
                    raise ServicePackageError from e
                if str(e) == "Parent module '' not loaded, cannot perform relative import":
                    logging.getLogger('import').warning('Invalid service package/parent name, may conflict with Python internals: "{}" - change parent folder name'.format(file_path.rsplit('/', 2)[1]))
                    raise ServicePackageError from e
                raise e
        except (ImportError, ModuleNotFoundError) as e:  # noqa
            if file_name.endswith('.py.py'):
                return cls.import_service_file(file_name[:-3])
            if file_name.endswith('.py') and isinstance(e, ModuleNotFoundError):  # noqa
                return cls.import_service_file(file_name[:-3])
            if file_name.endswith('.py'):
                file_name = file_name[:-3]
            logging.getLogger('import').warning('Invalid service, unable to load service file "{}.py"'.format(file_name))
            raise e
        except OSError:
            if file_name.endswith('.py'):
                file_name = file_name[:-3]
            logging.getLogger('import').warning('Invalid service, no such service file "{}.py"'.format(file_name))
            sys.exit(2)
        except Exception as e:
            if file_name.endswith('.py'):
                file_name = file_name[:-3]
            logging.getLogger('import').warning('Unable to load service file "{}.py"'.format(file_name))
            logging.getLogger('import').warning('Error: {}'.format(e))
            raise e
        return service_import

    @classmethod
    def import_module(cls, file_name: str) -> ModuleType:
        cwd = os.getcwd()
        file_path = '{}/{}'.format(os.path.realpath(cwd), file_name)

        spec = importlib.util.spec_from_file_location(file_name, file_path)  # type: Any
        module_import = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module_import)

        return module_import



import re
from typing import List, Union


class RegexMissmatchException(Exception):
    def __init__(self, value: str, pattern: str) -> None:
        message = 'RegexMissmatchException: value "{}" does not match pattern "{}"'.format(value, pattern)
        super().__init__(message)


class TooSmallException(Exception):
    def __init__(self, obj: Union[List, str], min_length: int) -> None:
        message = 'TooSmallException: "{}" is not at least of length {}'.format(obj, min_length)
        super().__init__(message)


class TooLargeException(Exception):
    def __init__(self, obj: Union[List, str], max_length: int) -> None:
        message = 'TooLargeException: "{}" length is larger than {}'.format(obj, max_length)
        super().__init__(message)


def validate_field_regex(value: str, pattern_str: str) -> None:
    pattern = re.compile(pattern_str)
    if not pattern.match(value):
        raise RegexMissmatchException(value=value, pattern=pattern_str)


def validate_field_length(obj: Union[List, str], min_length: int = -1, max_length: int = -1) -> None:
    if min_length != -1:
        if len(obj) < min_length:
            raise TooSmallException(obj=obj, min_length=min_length)
    if max_length != -1:
        if len(obj) > max_length:
            raise TooLargeException(obj=obj, max_length=max_length)



# @generated by generate_proto_mypy_stubs.py.  Do not edit!
import google.protobuf.message
import typing

class Service(google.protobuf.message.Message):
    name = ... # type: typing.Text
    uuid = ... # type: typing.Text

    def __init__(self,
        name : typing.Optional[typing.Text] = None,
        uuid : typing.Optional[typing.Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> Service: ...
    def MergeFrom(self, other_msg: google.protobuf.message.Message) -> None: ...
    def CopyFrom(self, other_msg: google.protobuf.message.Message) -> None: ...

class Metadata(google.protobuf.message.Message):
    message_uuid = ... # type: typing.Text
    protocol_version = ... # type: typing.Text
    timestamp = ... # type: float
    topic = ... # type: typing.Text
    data_encoding = ... # type: typing.Text

    def __init__(self,
        message_uuid : typing.Optional[typing.Text] = None,
        protocol_version : typing.Optional[typing.Text] = None,
        timestamp : typing.Optional[float] = None,
        topic : typing.Optional[typing.Text] = None,
        data_encoding : typing.Optional[typing.Text] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> Metadata: ...
    def MergeFrom(self, other_msg: google.protobuf.message.Message) -> None: ...
    def CopyFrom(self, other_msg: google.protobuf.message.Message) -> None: ...

class SNSSQSMessage(google.protobuf.message.Message):
    data = ... # type: bytes

    @property
    def service(self) -> Service: ...

    @property
    def metadata(self) -> Metadata: ...

    def __init__(self,
        service : typing.Optional[Service] = None,
        metadata : typing.Optional[Metadata] = None,
        data : typing.Optional[bytes] = None,
        ) -> None: ...
    @classmethod
    def FromString(cls, s: bytes) -> SNSSQSMessage: ...
    def MergeFrom(self, other_msg: google.protobuf.message.Message) -> None: ...
    def CopyFrom(self, other_msg: google.protobuf.message.Message) -> None: ...




from tomodachi.__version__ import __version__ as __version__, __version_info__ as __version_info__
from tomodachi.invoker import decorator
from tomodachi.transport.amqp import amqp as amqp, amqp_publish as amqp_publish
from tomodachi.transport.aws_sns_sqs import aws_sns_sqs as aws_sns_sqs, aws_sns_sqs_publish as aws_sns_sqs_publish
from tomodachi.transport.http import HttpException as HttpException, Response as HttpResponse, http as http, http_error as http_error, http_static as http_static, websocket as websocket, get_http_response_status as get_http_response_status
from tomodachi.transport.schedule import daily as daily, heartbeat as heartbeat, hourly as hourly, minutely as minutely, monthly as monthly, schedule as schedule
from typing import Any, Optional

CLASS_ATTRIBUTE: str = ...

def service(cls: Any) -> Any: ...

class Service:
    TOMODACHI_SERVICE_CLASS: bool = ...
    log: Any = ...
    log_setup: Any = ...

def set_service(name: str, instance: Any) -> None: ...

def get_service(name: Optional[str] = None) -> Any: ...

def get_instance(name: Optional[str] = None) -> Any: ...



#!/usr/bin/env python
import os
import sys
import getopt
import logging
from typing import List, Optional
import tomodachi
from tomodachi.launcher import ServiceLauncher
from tomodachi.watcher import Watcher
from tomodachi.config import parse_config_files


try:
    if ModuleNotFoundError:
        pass
except Exception:
    class ModuleNotFoundError(ImportError):
        pass


class CLI:
    def help_command_usage(self) -> str:
        return ('Usage: tomodachi.py subcommand [options] [args]\n'
                '\n'
                'Options:\n'
                '  -h, --help             show this help message and exit\n'
                '  -v, --version          print tomodachi version\n'
                '  --dependency-versions  print versions of dependencies\n'
                '\n'
                'Available subcommands:\n'
                '  run <service ...> [-c <config-file ...>] [--production]\n'
                '  -c, --config <files>   use json configuration files\n'
                '  -l, --log <level>      specify log level\n'
                '  --production           disable restart on file changes\n'
                )

    def help_command(self) -> None:
        print(self.help_command_usage())
        sys.exit(2)

    def version_command(self) -> None:
        print('tomodachi/{}'.format(tomodachi.__version__))
        sys.exit(0)

    def dependency_versions_command(self) -> None:
        self.test_dependencies(fail_on_errors=False, output_versions=True)
        sys.exit(0)

    def test_dependencies(self, fail_on_errors: bool = True, output_versions: bool = False) -> None:
        errors = False

        try:
            import aiobotocore
            if output_versions:
                print('aiobotocore/{}'.format(aiobotocore.__version__))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            print('Dependency failure: aiobotocore failed to load (error: "{}")'.format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            print('Dependency failure: aiobotocore failed to load (error: "{}")'.format(str(e)))
            logging.exception('')
            print('')

        try:
            import botocore
            if output_versions:
                print('botocore/{}'.format(botocore.__version__))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            print('Dependency failure: botocore failed to load (error: "{}")'.format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            print('Dependency failure: botocore failed to load (error: "{}")'.format(str(e)))
            logging.exception('')
            print('')

        try:
            import aiohttp
            if output_versions:
                print('aiohttp/{}'.format(aiohttp.__version__))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            print('Dependency failure: aiohttp failed to load (error: "{}")'.format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            print('Dependency failure: aiohttp failed to load (error: "{}")'.format(str(e)))
            logging.exception('')
            print('')

        try:
            import aioamqp
            if output_versions:
                print('aioamqp/{}'.format(aioamqp.__version__))
        except ModuleNotFoundError as e:  # pragma: no cover
            errors = True
            print('Dependency failure: aioamqp failed to load (error: "{}")'.format(str(e)))
        except Exception as e:  # pragma: no cover
            errors = True
            print('Dependency failure: aioamqp failed to load (error: "{}")'.format(str(e)))
            logging.exception('')
            print('')

        try:
            # Optional
            import google.protobuf
            if output_versions:
                print('protobuf/{}'.format(google.protobuf.__version__))
        except ModuleNotFoundError as e:  # pragma: no cover
            pass
        except Exception as e:  # pragma: no cover
            pass

        if not errors:
            try:
                import tomodachi.invoker
                import tomodachi.helpers.logging
                import tomodachi.transport.amqp
                import tomodachi.transport.aws_sns_sqs
                import tomodachi.transport.http
                import tomodachi.transport.schedule
            except Exception as e:  # pragma: no cover
                errors = True
                print('Dependency failure: tomodachi essentials failed to load (error: "{}")'.format(str(e)))
                logging.exception('')
                print('')

        if errors:
            if fail_on_errors:
                logging.getLogger('exception').warning('Unable to initialize dependencies')
                logging.getLogger('exception').warning('Error: See above exceptions and traceback')
                sys.exit(1)
            else:
                print('There were errors - see above for exceptions and traceback')

    def run_command_usage(self) -> str:
        return 'Usage: tomodachi.py run <service ...> [-c <config-file ...>] [--production]'

    def run_command(self, args: List[str]) -> None:
        if len(args) == 0:
            print(self.run_command_usage())
        else:
            configuration = None
            log_level = logging.INFO

            if '-c' in args or '--config' in args:
                index = args.index('-c') if '-c' in args else args.index('--config')
                args.pop(index)

                config_files = []  # type: List[str]
                while len(args) > index and args[index][0] != '-':
                    value = args.pop(index)
                    if value not in config_files:
                        config_files.append(value)

                if not len(config_files):
                    print('Missing config file on command line')
                    sys.exit(2)

                try:
                    configuration = parse_config_files(config_files)
                except FileNotFoundError as e:
                    print('Invalid config file: {}'.format(str(e)))
                    sys.exit(2)
                except ValueError as e:
                    print('Invalid config file, invalid JSON format: {}'.format(str(e)))
                    sys.exit(2)

            if '--production' in args:
                index = args.index('--production')
                args.pop(index)
                watcher = None
            else:
                cwd = os.getcwd()
                root_directories = [os.getcwd()]
                for arg in set(args):
                    root_directories.append(os.path.dirname('{}/{}'.format(os.path.realpath(cwd), arg)))
                watcher = Watcher(root=root_directories, configuration=configuration)

            if '-l' in args or '--log' in args:
                index = args.index('-l') if '-l' in args else args.index('--log')
                args.pop(index)
                if len(args) > index:
                    log_level = getattr(logging, args.pop(index).upper(), None) or log_level

            logging.basicConfig(format='%(asctime)s (%(name)s): %(message)s', level=log_level)
            logging.Formatter(fmt='%(asctime)s.%(msecs).03d', datefmt='%Y-%m-%d %H:%M:%S')

            self.test_dependencies()

            ServiceLauncher.run_until_complete(set(args), configuration, watcher)
        sys.exit(0)

    def main(self, argv: List[str]) -> None:
        try:
            opts, args = getopt.getopt(argv, "hlvV ", ['help', 'log', 'version', 'version', 'dependency-versions'])
        except getopt.GetoptError:
            self.help_command()
        for opt, _ in opts:
            if opt in ['-h', '--help']:
                self.help_command()
            if opt in ['-v', '-V', '--version']:
                self.version_command()
            if opt in ['--dependency-versions']:
                self.dependency_versions_command()
        if len(args):
            if args[0] in ('run', 'start', 'go'):
                self.run_command(args[1:])
        self.help_command()


def cli_entrypoint(argv: Optional[List[str]] = None) -> None:
    if argv is None:
        argv = sys.argv
    CLI().main(argv[1:])



import logging
import asyncio
import aiobotocore
import botocore
import aiohttp
import time
import hashlib
import re
import binascii
import ujson
import uuid
import inspect
import functools
from botocore.parsers import ResponseParserError
from typing import Any, Dict, Union, Optional, Callable, List, Tuple, Match, Awaitable
from tomodachi.invoker import Invoker
from tomodachi.helpers.dict import merge_dicts
from tomodachi.helpers.middleware import execute_middlewares

DRAIN_MESSAGE_PAYLOAD = '__TOMODACHI_DRAIN__cdab4416-1727-4603-87c9-0ff8dddf1f22__'
MESSAGE_PROTOCOL_DEFAULT = 'e6fb6007-cf15-4cfd-af2e-1d1683374e70'


class AWSSNSSQSException(Exception):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._log_level = kwargs.get('log_level') if kwargs and kwargs.get('log_level') else 'INFO'


class AWSSNSSQSConnectionException(AWSSNSSQSException):
    pass


class AWSSNSSQSInternalServiceError(AWSSNSSQSException):
    pass


class AWSSNSSQSInternalServiceErrorException(AWSSNSSQSInternalServiceError):
    pass


class AWSSNSSQSInternalServiceException(AWSSNSSQSInternalServiceError):
    pass


class AWSSNSSQSTransport(Invoker):
    clients = None
    clients_creation_time = None
    topics = {}  # type: Dict[str, str]
    close_waiter = None

    @classmethod
    async def publish(cls, service: Any, data: Any, topic: str, wait: bool = True, message_protocol: Any = MESSAGE_PROTOCOL_DEFAULT, **kwargs: Any) -> None:
        message_protocol = getattr(service, 'message_protocol', None) if message_protocol == MESSAGE_PROTOCOL_DEFAULT else message_protocol

        payload = data
        if message_protocol:
            build_message_func = getattr(message_protocol, 'build_message', None)
            if build_message_func:
                payload = await build_message_func(service, topic, data, **kwargs)

        topic_arn = await cls.create_topic(cls, topic, service.context)

        async def _publish_message() -> None:
            await cls.publish_message(cls, topic_arn, payload, service.context)

        if wait:
            await _publish_message()
        else:
            loop = asyncio.get_event_loop()  # type: Any
            loop.create_task(_publish_message())

    @classmethod
    def get_topic_name(cls, topic: str, context: Dict) -> str:
        if context.get('options', {}).get('aws_sns_sqs', {}).get('topic_prefix'):
            return '{}{}'.format(context.get('options', {}).get('aws_sns_sqs', {}).get('topic_prefix'), topic)
        return topic

    @classmethod
    def get_topic_name_without_prefix(cls, topic: str, context: Dict) -> str:
        if context.get('options', {}).get('aws_sns_sqs', {}).get('topic_prefix'):
            if topic.startswith(context.get('options', {}).get('aws_sns_sqs', {}).get('topic_prefix')):
                prefix_length = len(context.get('options', {}).get('aws_sns_sqs', {}).get('topic_prefix', ''))
                return topic[prefix_length:]
        return topic

    @classmethod
    def get_topic_from_arn(cls, topic: str) -> str:
        return topic.rsplit(':')[-1]

    @classmethod
    def decode_topic(cls, encoded_topic: str) -> str:
        def decode(match: Match) -> str:
            return binascii.unhexlify(match.group(1).encode('utf-8')).decode('utf-8')

        return re.sub(r'___([a-f0-9]{2}|[a-f0-9]{4}|[a-f0-9]{6}|[a-f0-9]{8})_', decode, encoded_topic)

    @classmethod
    def encode_topic(cls, topic: str) -> str:
        def encode(match: Match) -> str:
            return '___' + binascii.hexlify(match.group(1).encode('utf-8')).decode('utf-8') + '_'

        return re.sub(r'([^a-zA-Z0-9_*#-])', encode, topic)

    @classmethod
    def get_queue_name(cls, topic: str, func_name: str, _uuid: str, competing_consumer: bool, context: Dict) -> str:
        if not competing_consumer:
            queue_name = hashlib.sha256('{}{}{}'.format(topic, func_name, _uuid).encode('utf-8')).hexdigest()
        else:
            queue_name = hashlib.sha256(topic.encode('utf-8')).hexdigest()

        if context.get('options', {}).get('aws_sns_sqs', {}).get('queue_name_prefix'):
            return '{}{}'.format(context.get('options', {}).get('aws_sns_sqs', {}).get('queue_name_prefix'), queue_name)
        return queue_name

    @classmethod
    def prefix_queue_name(cls, queue_name: str, context: Dict) -> str:
        if context.get('options', {}).get('aws_sns_sqs', {}).get('queue_name_prefix'):
            return '{}{}'.format(context.get('options', {}).get('aws_sns_sqs', {}).get('queue_name_prefix'), queue_name)
        return queue_name

    async def subscribe_handler(cls: Any, obj: Any, context: Dict, func: Any, topic: str, callback_kwargs: Optional[Union[list, set, tuple]] = None, competing: Optional[bool] = None, queue_name: Optional[str] = None, message_protocol: Any = MESSAGE_PROTOCOL_DEFAULT, **kwargs: Any) -> Any:
        parser_kwargs = kwargs
        message_protocol = context.get('message_protocol') if message_protocol == MESSAGE_PROTOCOL_DEFAULT else message_protocol

        # Validate the parser kwargs if there is a validation function in the protocol
        if message_protocol:
            protocol_kwargs_validation_func = getattr(message_protocol, 'validate', None)
            if protocol_kwargs_validation_func:
                protocol_kwargs_validation_func(**parser_kwargs)

        async def handler(payload: Optional[str], receipt_handle: Optional[str] = None, queue_url: Optional[str] = None, message_topic: str = '') -> Any:
            if not payload or payload == DRAIN_MESSAGE_PAYLOAD:
                await cls.delete_message(cls, receipt_handle, queue_url, context)
                return

            _callback_kwargs = callback_kwargs  # type: Any
            values = inspect.getfullargspec(func)
            if not _callback_kwargs:
                _callback_kwargs = {k: values.defaults[i - len(values.args) + 1] if values.defaults and i >= len(values.args) - len(values.defaults) - 1 else None for i, k in enumerate(values.args[1:])} if values.args and len(values.args) > 1 else {}
            else:
                _callback_kwargs = {k: None for k in _callback_kwargs if k != 'self'}
            kwargs = {k: v for k, v in _callback_kwargs.items()}

            message = payload
            message_uuid = None
            message_key = None
            if message_protocol:
                try:
                    parse_message_func = getattr(message_protocol, 'parse_message', None)
                    if parse_message_func:
                        if len(parser_kwargs):
                            message, message_uuid, timestamp = await parse_message_func(payload, **parser_kwargs)
                        else:
                            message, message_uuid, timestamp = await parse_message_func(payload)
                    if message is not False and message_uuid:
                        if not context.get('_aws_sns_sqs_received_messages'):
                            context['_aws_sns_sqs_received_messages'] = {}
                        message_key = '{}:{}'.format(message_uuid, func.__name__)
                        if context['_aws_sns_sqs_received_messages'].get(message_key):
                            return
                        context['_aws_sns_sqs_received_messages'][message_key] = time.time()
                        _received_messages = context['_aws_sns_sqs_received_messages']
                        if _received_messages and isinstance(_received_messages, dict) and len(_received_messages) > 100000:
                            context['_aws_sns_sqs_received_messages'] = {k: v for k, v in context['_aws_sns_sqs_received_messages'].items() if v > time.time() - 60}

                    if _callback_kwargs:
                        if isinstance(message, dict):
                            for k, v in message.items():
                                if k in _callback_kwargs:
                                    kwargs[k] = v
                        if 'message' in _callback_kwargs and 'message' not in message:
                            kwargs['message'] = message
                except Exception as e:
                    logging.getLogger('exception').exception('Uncaught exception: {}'.format(str(e)))
                    if message is not False and not message_uuid:
                        await cls.delete_message(cls, receipt_handle, queue_url, context)
                    elif message is False and message_uuid:
                        pass  # incompatible protocol, should probably delete if old message
                    elif message is False:
                        await cls.delete_message(cls, receipt_handle, queue_url, context)
                    return

            @functools.wraps(func)
            async def routine_func(*a: Any, **kw: Any) -> Any:
                try:
                    if not message_protocol and len(values.args[1:]):
                        routine = func(*(obj, message, *a))
                    elif len(merge_dicts(kwargs, kw)):
                        routine = func(*(obj, *a), **merge_dicts(kwargs, kw))
                    elif len(values.args[1:]):
                        routine = func(*(obj, message, *a), **kw)
                    else:
                        routine = func(*(obj, *a), **kw)
                except Exception as e:
                    logging.getLogger('exception').exception('Uncaught exception: {}'.format(str(e)))
                    if issubclass(e.__class__, (AWSSNSSQSInternalServiceError, AWSSNSSQSInternalServiceErrorException, AWSSNSSQSInternalServiceException)):
                        if message_key:
                            del context['_aws_sns_sqs_received_messages'][message_key]
                        return
                    await cls.delete_message(cls, receipt_handle, queue_url, context)
                    return

                if isinstance(routine, Awaitable):
                    try:
                        return_value = await routine
                    except Exception as e:
                        logging.getLogger('exception').exception('Uncaught exception: {}'.format(str(e)))
                        if issubclass(e.__class__, (AWSSNSSQSInternalServiceError, AWSSNSSQSInternalServiceErrorException, AWSSNSSQSInternalServiceException)):
                            if message_key:
                                del context['_aws_sns_sqs_received_messages'][message_key]
                            return
                        await cls.delete_message(cls, receipt_handle, queue_url, context)
                        return
                else:
                    return_value = routine

                await cls.delete_message(cls, receipt_handle, queue_url, context)
                return return_value

            return_value = await execute_middlewares(func, routine_func, context.get('message_middleware', []), *(obj, message, topic))
            return return_value

        context['_aws_sns_sqs_subscribers'] = context.get('_aws_sns_sqs_subscribers', [])
        context['_aws_sns_sqs_subscribers'].append((topic, competing, queue_name, func, handler))

        start_func = cls.subscribe(cls, obj, context)
        return (await start_func) if start_func else None

    def create_client(cls: Any, name: str, context: Dict) -> None:
        logging.getLogger('botocore.vendored.requests.packages.urllib3.connectionpool').setLevel(logging.WARNING)

        if not cls.clients:
            cls.clients = {}
            cls.clients_creation_time = {}
        loop = asyncio.get_event_loop()
        session = aiobotocore.get_session(loop=loop)

        config_base = context.get('options', {}).get('aws_sns_sqs', context.get('options', {}).get('aws', {}))
        aws_config_base = context.get('options', {}).get('aws', {})

        region_name = config_base.get('aws_region_name', config_base.get('region_name')) or \
            aws_config_base.get('aws_region_name', aws_config_base.get('region_name'))
        aws_secret_access_key = config_base.get('aws_secret_access_key', config_base.get('secret_access_key')) or \
            aws_config_base.get('aws_secret_access_key', aws_config_base.get('secret_access_key'))
        aws_access_key_id = config_base.get('aws_access_key_id', config_base.get('access_key_id')) or \
            aws_config_base.get('aws_access_key_id', aws_config_base.get('access_key_id'))
        endpoint_url = config_base.get('aws_endpoint_urls', config_base.get('endpoint_urls', {})).get(name) or \
            config_base.get('aws_{}_endpoint_url'.format(name), config_base.get('{}_endpoint_url'.format(name))) or \
            aws_config_base.get('aws_endpoint_urls', aws_config_base.get('endpoint_urls', {})).get(name) or \
            config_base.get('aws_endpoint_url', config_base.get('endpoint_url')) or \
            aws_config_base.get('aws_endpoint_url', aws_config_base.get('endpoint_url')) or \
            context.get('options', {}).get('aws_endpoint_urls', {}).get(name)

        try:
            if cls.clients_creation_time.get(name) and cls.clients_creation_time[name] + 30 > time.time():
                return
            cls.clients[name] = session.create_client(name, region_name=region_name, aws_secret_access_key=aws_secret_access_key, aws_access_key_id=aws_access_key_id, endpoint_url=endpoint_url)
            cls.clients_creation_time[name] = time.time()
        except (botocore.exceptions.PartialCredentialsError, botocore.exceptions.NoRegionError) as e:
            error_message = str(e)
            logging.getLogger('transport.aws_sns_sqs').warning('Invalid credentials [{}] to AWS ({})'.format(name, error_message))
            raise AWSSNSSQSConnectionException(error_message, log_level=context.get('log_level')) from e

    async def create_topic(cls: Any, topic: str, context: Dict) -> str:
        if not cls.topics:
            cls.topics = {}
        if cls.topics.get(topic):
            topic_arn = cls.topics.get(topic)
            if topic_arn and isinstance(topic_arn, str):
                return topic_arn

        if not cls.clients or not cls.clients.get('sns'):
            cls.create_client(cls, 'sns', context)
        client = cls.clients.get('sns')

        try:
            response = await asyncio.wait_for(client.create_topic(Name=cls.encode_topic(cls.get_topic_name(topic, context))), timeout=30)
        except (botocore.exceptions.NoCredentialsError, aiohttp.client_exceptions.ClientOSError) as e:
            error_message = str(e)
            logging.getLogger('transport.aws_sns_sqs').warning('Unable to connect [sns] to AWS ({})'.format(error_message))
            raise AWSSNSSQSConnectionException(error_message, log_level=context.get('log_level')) from e
        except (botocore.exceptions.PartialCredentialsError, botocore.exceptions.ClientError, asyncio.TimeoutError) as e:
            error_message = str(e) if not isinstance(e, asyncio.TimeoutError) else 'Network timeout'
            logging.getLogger('transport.aws_sns_sqs').warning('Unable to create topic [sns] on AWS ({})'.format(error_message))
            raise AWSSNSSQSException(error_message, log_level=context.get('log_level')) from e

        topic_arn = response.get('TopicArn')
        if not topic_arn or not isinstance(topic_arn, str):
            error_message = 'Missing ARN in response'
            logging.getLogger('transport.aws_sns_sqs').warning('Unable to create topic [sns] on AWS ({})'.format(error_message))
            raise AWSSNSSQSException(error_message, log_level=context.get('log_level'))

        cls.topics[topic] = topic_arn

        return topic_arn

    async def publish_message(cls: Any, topic_arn: str, message: Any, context: Dict) -> str:
        if not cls.clients or not cls.clients.get('sns'):
            cls.create_client(cls, 'sns', context)

        for retry in range(1, 4):
            client = cls.clients.get('sns')
            try:
                response = await asyncio.wait_for(client.publish(TopicArn=topic_arn, Message=message), timeout=30)
            except (aiohttp.client_exceptions.ServerDisconnectedError, RuntimeError) as e:
                await asyncio.sleep(1)
                try:
                    task = client.close()
                    if getattr(task, '_coro', None):
                        task = task._coro
                    await asyncio.wait([asyncio.ensure_future(task)], timeout=3)
                except Exception:
                    pass
                cls.create_client(cls, 'sns', context)
                if retry >= 3:
                    raise e
                continue
            except (botocore.exceptions.ClientError, aiohttp.client_exceptions.ClientConnectorError, asyncio.TimeoutError) as e:
                if retry >= 3:
                    error_message = str(e) if not isinstance(e, asyncio.TimeoutError) else 'Network timeout'
                    logging.getLogger('transport.aws_sns_sqs').warning('Unable to publish message [sns] on AWS ({})'.format(error_message))
                    raise AWSSNSSQSException(error_message, log_level=context.get('log_level')) from e
                continue
            break

        message_id = response.get('MessageId')
        if not message_id or not isinstance(message_id, str):
            error_message = 'Missing MessageId in response'
            logging.getLogger('transport.aws_sns_sqs').warning('Unable to publish message [sns] on AWS ({})'.format(error_message))
            raise AWSSNSSQSException(error_message, log_level=context.get('log_level'))

        return message_id

    async def delete_message(cls: Any, receipt_handle: Optional[str], queue_url: Optional[str], context: Dict) -> None:
        if not receipt_handle:
            return
        if not cls.clients or not cls.clients.get('sqs'):
            cls.create_client(cls, 'sqs', context)

        async def _delete_message() -> None:
            for retry in range(1, 4):
                client = cls.clients.get('sqs')
                try:
                    await asyncio.wait_for(client.delete_message(ReceiptHandle=receipt_handle, QueueUrl=queue_url), timeout=30)
                except (aiohttp.client_exceptions.ServerDisconnectedError, RuntimeError) as e:
                    await asyncio.sleep(1)
                    try:
                        task = client.close()
                        if getattr(task, '_coro', None):
                            task = task._coro
                        await asyncio.wait([asyncio.ensure_future(task)], timeout=3)
                    except Exception:
                        pass
                    cls.create_client(cls, 'sqs', context)
                    if retry >= 3:
                        raise e
                    continue
                except botocore.exceptions.ClientError as e:
                    error_message = str(e)
                    logging.getLogger('transport.aws_sns_sqs').warning('Unable to delete message [sqs] on AWS ({})'.format(error_message))
                except asyncio.TimeoutError as e:
                    if retry >= 3:
                        error_message = 'Network timeout'
                        logging.getLogger('transport.aws_sns_sqs').warning('Unable to delete message [sqs] on AWS ({})'.format(error_message))
                        raise AWSSNSSQSException(error_message, log_level=context.get('log_level')) from e
                    continue

        await _delete_message()

    async def create_queue(cls: Any, queue_name: str, context: Dict) -> Tuple[str, str]:
        if not cls.clients or not cls.clients.get('sqs'):
            cls.create_client(cls, 'sqs', context)
        client = cls.clients.get('sqs')

        try:
            response = await client.create_queue(QueueName=queue_name)
        except (botocore.exceptions.NoCredentialsError, botocore.exceptions.PartialCredentialsError, aiohttp.client_exceptions.ClientOSError) as e:
            error_message = str(e)
            logging.getLogger('transport.aws_sns_sqs').warning('Unable to connect [sqs] to AWS ({})'.format(error_message))
            raise AWSSNSSQSConnectionException(error_message, log_level=context.get('log_level')) from e
        except botocore.exceptions.ClientError as e:
            error_message = str(e)
            logging.getLogger('transport.aws_sns_sqs').warning('Unable to create queue [sqs] on AWS ({})'.format(error_message))
            raise AWSSNSSQSException(error_message, log_level=context.get('log_level')) from e

        queue_url = response.get('QueueUrl')
        if not queue_url:
            error_message = 'Missing Queue URL in response'
            logging.getLogger('transport.aws_sns_sqs').warning('Unable to create queue [sqs] on AWS ({})'.format(error_message))
            raise AWSSNSSQSException(error_message, log_level=context.get('log_level'))

        try:
            response = await client.get_queue_attributes(QueueUrl=queue_url, AttributeNames=['QueueArn'])
        except botocore.exceptions.ClientError as e:
            error_message = str(e)
            logging.getLogger('transport.aws_sns_sqs').warning('Unable to get queue attributes [sqs] on AWS ({})'.format(error_message))
            raise AWSSNSSQSException(error_message, log_level=context.get('log_level')) from e

        queue_arn = response.get('Attributes', {}).get('QueueArn')
        if not queue_arn:
            error_message = 'Missing ARN in response'
            logging.getLogger('transport.aws_sns_sqs').warning('Unable to get queue attributes [sqs] on AWS ({})'.format(error_message))
            raise AWSSNSSQSException(error_message, log_level=context.get('log_level'))

        return queue_url, queue_arn

    @classmethod
    def generate_queue_policy(cls, queue_arn: str, topic_arn_list: List, context: Dict) -> Dict:
        if len(topic_arn_list) == 1:
            if context.get('options', {}).get('aws_sns_sqs', {}).get('queue_policy'):
                source_arn = context.get('options', {}).get('aws_sns_sqs', {}).get('queue_policy')
            else:
                source_arn = topic_arn_list[0]
        else:
            wildcard_topic_arn = []
            try:
                for i in range(0, min([len(topic_arn) for topic_arn in topic_arn_list])):
                    if len(set([topic_arn[i] for topic_arn in topic_arn_list])) == 1:
                        wildcard_topic_arn.append(topic_arn_list[0][i])
                    else:
                        wildcard_topic_arn.append('*')
                        break
            except IndexError:
                wildcard_topic_arn.append('*')

            source_arn = ''.join(wildcard_topic_arn)
            if context.get('options', {}).get('aws_sns_sqs', {}).get('queue_policy'):
                source_arn = context.get('options', {}).get('aws_sns_sqs', {}).get('queue_policy')
            if context.get('options', {}).get('aws_sns_sqs', {}).get('wildcard_queue_policy'):
                source_arn = context.get('options', {}).get('aws_sns_sqs', {}).get('wildcard_queue_policy')

        queue_policy = {
            "Version": "2012-10-17",
            "Id": "{}/SQSDefaultPolicy".format(queue_arn),
            "Statement": [
                {
                    "Sid": "{}{}".format(str(uuid.uuid4()), ('%.3f' % time.time()).replace('.', '')),
                    "Effect": "Allow",
                    "Principal": "*",
                    "Action": "SQS:SendMessage",
                    "Resource": queue_arn,
                    "Condition": {
                        "ArnEquals": {
                            "aws:SourceArn": source_arn
                        }
                    }
                }
            ]
        }
        return queue_policy

    async def subscribe_wildcard_topic(cls: Any, topic: str, queue_arn: str, queue_url: str, context: Dict) -> Optional[List]:
        if not cls.clients or not cls.clients.get('sns'):
            cls.create_client(cls, 'sns', context)
        client = cls.clients.get('sns')

        pattern = r'^arn:aws:sns:[^:]+:[^:]+:{}$'.format(cls.encode_topic(cls.get_topic_name(topic, context)).replace(cls.encode_topic('*'), '((?!{}).)*'.format(cls.encode_topic('.'))).replace(cls.encode_topic('#'), '.*'))
        compiled_pattern = re.compile(pattern)

        next_token = False
        topic_arn_list = None
        while next_token is not None:
            try:
                if next_token:
                    response = await client.list_topics(NextToken=next_token)
                else:
                    response = await client.list_topics()
            except botocore.exceptions.ClientError as e:
                error_message = str(e)
                logging.getLogger('transport.aws_sns_sqs').warning('Unable to list topics [sns] on AWS ({})'.format(error_message))
                raise AWSSNSSQSException(error_message, log_level=context.get('log_level')) from e

            next_token = response.get('NextToken')
            topics = response.get('Topics', [])
            topic_arn_list = [t.get('TopicArn') for t in topics if t.get('TopicArn') and compiled_pattern.match(t.get('TopicArn'))]

        if topic_arn_list:
            queue_policy = cls.generate_queue_policy(queue_arn, topic_arn_list, context)
            cls.topics[topic] = topic_arn_list[0]
            return await cls.subscribe_topics(cls, topic_arn_list, queue_arn, queue_url, context, queue_policy=queue_policy)

        return None

    async def subscribe_topics(cls: Any, topic_arn_list: List, queue_arn: str, queue_url: str, context: Dict, queue_policy: Optional[Dict] = None) -> List:
        if not cls.clients or not cls.clients.get('sns'):
            cls.create_client(cls, 'sns', context)
        client = cls.clients.get('sns')

        if not cls.clients or not cls.clients.get('sqs'):
            cls.create_client(cls, 'sqs', context)
        sqs_client = cls.clients.get('sqs')

        if not queue_policy:
            queue_policy = cls.generate_queue_policy(queue_arn, topic_arn_list, context)

        try:
            # MessageRetentionPeriod (default 4 days, set to context value)
            # VisibilityTimeout (default 30 seconds)
            response = await sqs_client.set_queue_attributes(QueueUrl=queue_url, Attributes={'Policy': ujson.dumps(queue_policy)})
        except botocore.exceptions.ClientError as e:
            error_message = str(e)
            logging.getLogger('transport.aws_sns_sqs').warning('Unable to set queue attributes [sqs] on AWS ({})'.format(error_message))
            raise AWSSNSSQSException(error_message, log_level=context.get('log_level')) from e

        subscription_arn_list = []
        for topic_arn in topic_arn_list:
            try:
                response = await client.subscribe(TopicArn=topic_arn, Protocol='sqs', Endpoint=queue_arn)
            except botocore.exceptions.ClientError as e:
                error_message = str(e)
                logging.getLogger('transport.aws_sns_sqs').warning('Unable to subscribe to topic [sns] on AWS ({})'.format(error_message))
                raise AWSSNSSQSException(error_message, log_level=context.get('log_level')) from e

            subscription_arn = response.get('SubscriptionArn')
            if not subscription_arn:
                error_message = 'Missing Subscription ARN in response'
                logging.getLogger('transport.aws_sns_sqs').warning('Unable to subscribe to topic [sns] on AWS ({})'.format(error_message))
                raise AWSSNSSQSException(error_message, log_level=context.get('log_level'))

            subscription_arn_list.append(subscription_arn)

        return subscription_arn_list

    async def consume_queue(cls: Any, obj: Any, context: Dict, handler: Callable, queue_url: str) -> None:
        if not cls.clients or not cls.clients.get('sqs'):
            cls.create_client(cls, 'sqs', context)

        if not cls.close_waiter:
            cls.close_waiter = asyncio.Future()
        stop_waiter = asyncio.Future()  # type: asyncio.Future
        start_waiter = asyncio.Future()  # type: asyncio.Future

        async def receive_messages() -> None:
            def callback(payload: Optional[str], receipt_handle: Optional[str], queue_url: Optional[str], message_topic: str) -> Callable:
                async def _callback() -> None:
                    await handler(payload, receipt_handle, queue_url, message_topic)
                return _callback

            client = cls.clients.get('sqs')

            await start_waiter
            is_disconnected = False
            while not cls.close_waiter.done():
                try:
                    response = await asyncio.wait_for(client.receive_message(QueueUrl=queue_url, WaitTimeSeconds=20, MaxNumberOfMessages=10), timeout=30)
                    if is_disconnected:
                        is_disconnected = False
                        logging.getLogger('transport.aws_sns_sqs').warning('Reconnected - receiving messages')
                except (aiohttp.client_exceptions.ServerDisconnectedError, RuntimeError) as e:
                    if not is_disconnected:
                        is_disconnected = True
                        error_message = str(e) if e and str(e) not in ['', 'None'] else 'Server disconnected'
                        logging.getLogger('transport.aws_sns_sqs').warning('Unable to receive message from queue [sqs] on AWS ({}) - reconnecting'.format(error_message))
                    await asyncio.sleep(1)
                    try:
                        task = client.close()
                        if getattr(task, '_coro', None):
                            task = task._coro
                        await asyncio.wait([asyncio.ensure_future(task)], timeout=3)
                    except Exception:
                        pass
                    cls.create_client(cls, 'sqs', context)
                    client = cls.clients.get('sqs')
                    continue
                except ResponseParserError as e:
                    if not is_disconnected:
                        is_disconnected = True
                        error_message = 'Unable to parse response: the server was not able to produce a timely response to your request'
                        logging.getLogger('transport.aws_sns_sqs').warning('Unable to receive message from queue [sqs] on AWS ({}) - reconnecting'.format(error_message))
                    await asyncio.sleep(1)
                    continue
                except (botocore.exceptions.ClientError, aiohttp.client_exceptions.ClientConnectorError, asyncio.TimeoutError) as e:
                    error_message = str(e) if not isinstance(e, asyncio.TimeoutError) else 'Network timeout'
                    if 'AWS.SimpleQueueService.NonExistentQueue' in error_message:
                        if is_disconnected:
                            is_disconnected = False
                            logging.getLogger('transport.aws_sns_sqs').warning('Reconnected - receiving messages')
                        try:
                            context['_aws_sns_sqs_subscribed'] = False
                            cls.topics = {}
                            func = await cls.subscribe(cls, obj, context)
                            await func()
                        except Exception:
                            pass
                        await asyncio.sleep(20)
                        continue
                    if not is_disconnected:
                        logging.getLogger('transport.aws_sns_sqs').warning('Unable to receive message from queue [sqs] on AWS ({})'.format(error_message))
                    if isinstance(e, (asyncio.TimeoutError, aiohttp.client_exceptions.ClientConnectorError)):
                        is_disconnected = True
                    await asyncio.sleep(1)
                    continue
                except Exception as e:
                    error_message = str(e)
                    logging.getLogger('transport.aws_sns_sqs').warning('Unexpected error while receiving message from queue [sqs] on AWS ({})'.format(error_message))
                    await asyncio.sleep(1)
                    continue

                messages = response.get('Messages', [])
                if messages:
                    futures = []

                    for message in messages:
                        receipt_handle = message.get('ReceiptHandle')
                        try:
                            message_body = ujson.loads(message.get('Body'))
                            message_topic = cls.get_topic_name_without_prefix(cls.decode_topic(cls.get_topic_from_arn(message_body.get('TopicArn'))), context) if message_body.get('TopicArn') else ''
                        except ValueError:
                            # Malformed SQS message, not in SNS format and should be discarded
                            await cls.delete_message(cls, receipt_handle, queue_url, context)
                            logging.getLogger('transport.aws_sns_sqs').warning('Discarded malformed message')
                            continue

                        payload = message_body.get('Message')
                        futures.append(callback(payload, receipt_handle, queue_url, message_topic))
                    if futures:
                        await asyncio.wait([asyncio.ensure_future(asyncio.shield(func())) for func in futures if func])

            if not stop_waiter.done():
                stop_waiter.set_result(None)

        loop = asyncio.get_event_loop()  # type: Any

        stop_method = getattr(obj, '_stop_service', None)

        async def stop_service(*args: Any, **kwargs: Any) -> None:
            if not cls.close_waiter.done():
                cls.close_waiter.set_result(None)
                logging.getLogger('transport.aws_sns_sqs').warning('Draining message pool - may take up to 30 seconds')

                master_stop_time = time.time()
                for _ in range(0, 100):
                    tasks = []
                    t = time.time()
                    if stop_waiter.done():
                        break
                    try:
                        for topic, topic_arn in cls.topics.items():
                            tasks.append(asyncio.ensure_future(cls.publish_message(cls, topic_arn, DRAIN_MESSAGE_PAYLOAD, context)))
                        if tasks:
                            task_results = await asyncio.wait(tasks, timeout=35)
                            exception = None
                            for v in [value for value in task_results][0]:
                                try:
                                    if (v.exception() or v.cancelled()) and not exception:
                                        exception = True
                                        sleep_timer = (t + 30.0) - time.time()
                                        if sleep_timer > 0.0:
                                            logging.getLogger('transport.aws_sns_sqs').warning('Draining failed - please wait')
                                            await asyncio.sleep(sleep_timer)
                                except Exception:
                                    pass
                            if exception and not stop_waiter.done():
                                stop_waiter.set_result(None)
                    except Exception as e:
                        pass

                    if master_stop_time + 35.0 > time.time() and not stop_waiter.done():
                        stop_waiter.set_result(None)

                    await asyncio.sleep(0.2)

                if not start_waiter.done():
                    start_waiter.set_result(None)
                await stop_waiter
                if stop_method:
                    await stop_method(*args, **kwargs)
                tasks = []
                for _, client in cls.clients.items():
                    task = client.close()
                    if getattr(task, '_coro', None):
                        task = task._coro
                    tasks.append(asyncio.ensure_future(task))
                await asyncio.wait(tasks, timeout=3)
                cls.clients = None
            else:
                await stop_waiter
                if stop_method:
                    await stop_method(*args, **kwargs)

        setattr(obj, '_stop_service', stop_service)

        started_method = getattr(obj, '_started_service', None)

        async def started_service(*args: Any, **kwargs: Any) -> None:
            if started_method:
                await started_method(*args, **kwargs)
            if not start_waiter.done():
                start_waiter.set_result(None)

        setattr(obj, '_started_service', started_service)

        loop.create_task(receive_messages())

    async def subscribe(cls: Any, obj: Any, context: Dict) -> Optional[Callable]:
        if context.get('_aws_sns_sqs_subscribed'):
            return None
        context['_aws_sns_sqs_subscribed'] = True

        if cls.clients:
            tasks = []
            for _, client in cls.clients.items():
                task = client.close()
                if getattr(task, '_coro', None):
                    task = task._coro
                tasks.append(asyncio.ensure_future(task))
            await asyncio.wait(tasks, timeout=3)
            cls.clients = None

        async def _subscribe() -> None:
            cls.close_waiter = asyncio.Future()

            async def setup_queue(topic: str, func: Callable, queue_name: Optional[str] = None, competing_consumer: Optional[bool] = None) -> str:
                _uuid = obj.uuid

                if queue_name and competing_consumer is False:
                    raise AWSSNSSQSException('Queue with predefined queue name must be competing', log_level=context.get('log_level'))

                if queue_name is None:
                    queue_name = cls.get_queue_name(cls.encode_topic(topic), func.__name__, _uuid, competing_consumer, context)
                else:
                    queue_name = cls.prefix_queue_name(queue_name, context)

                queue_url, queue_arn = await cls.create_queue(cls, queue_name, context)  # type: str, str

                if re.search(r'([*#])', topic):
                    await cls.subscribe_wildcard_topic(cls, topic, queue_arn, queue_url, context)
                else:
                    topic_arn = await cls.create_topic(cls, topic, context)
                    await cls.subscribe_topics(cls, (topic_arn,), queue_arn, queue_url, context)

                return queue_url

            for topic, competing, queue_name, func, handler in context.get('_aws_sns_sqs_subscribers', []):
                queue_url = await setup_queue(topic, func, competing_consumer=competing, queue_name=queue_name)
                await cls.consume_queue(cls, obj, context, handler, queue_url=queue_url)

        return _subscribe


aws_sns_sqs = AWSSNSSQSTransport.decorator(AWSSNSSQSTransport.subscribe_handler)
aws_sns_sqs_publish = AWSSNSSQSTransport.publish
publish = AWSSNSSQSTransport.publish



__version_info__ = (0, 16, 0)
__version__ = ''.join(['.{}'.format(str(n)) if type(n) is int else str(n) for n in __version_info__]).replace('.', '', 1 if type(__version_info__[0]) is int else 0)

if __name__ == "__main__":
    print(__version__)



import logging
import uuid
import time
import base64
import zlib
from typing import Any, Dict, Tuple, Union

from tomodachi.protocol.proto_build.protobuf.sns_sqs_message_pb2 import SNSSQSMessage

PROTOCOL_VERSION = 'tomodachi-protobuf-base--1.0.0'


class ProtobufBase(object):
    @classmethod
    def validate(cls, **kwargs: Any) -> None:
        if 'proto_class' not in kwargs:
            raise Exception('No proto_class defined')
        if kwargs.get('proto_class', None).__class__.__name__ != 'GeneratedProtocolMessageType':
            raise Exception('proto_class is not a GeneratedProtocolMessageType')

    @classmethod
    async def build_message(cls, service: Any, topic: str, data: Any, **kwargs: Any) -> str:
        message_data = data.SerializeToString()

        data_encoding = 'proto'
        if len(message_data) > 60000:
            message_data = zlib.compress(data.SerializeToString())
            data_encoding = 'gzip_proto'

        message = SNSSQSMessage()
        message.service.name = getattr(service, 'name', None)
        message.service.uuid = getattr(service, 'uuid', None)
        message.metadata.message_uuid = '{}.{}'.format(getattr(service, 'uuid', ''), str(uuid.uuid4()))
        message.metadata.protocol_version = PROTOCOL_VERSION
        message.metadata.timestamp = time.time()
        message.metadata.topic = topic
        message.metadata.data_encoding = data_encoding
        message.data = message_data

        return base64.b64encode(message.SerializeToString()).decode('ascii')

    @classmethod
    async def parse_message(cls, payload: str, proto_class: Any = None, validator: Any = None, **kwargs: Any) -> Union[Dict, Tuple]:
        message = SNSSQSMessage()
        message.ParseFromString(base64.b64decode(payload))

        message_uuid = message.metadata.message_uuid
        timestamp = message.metadata.timestamp

        raw_data = None
        obj = None

        if not proto_class:
            raw_data = message.data
        else:
            obj = proto_class()
            if message.metadata.data_encoding == 'proto':
                obj.ParseFromString(message.data)
            elif message.metadata.data_encoding == 'base64':  # deprecated
                obj.ParseFromString(base64.b64decode(message.data))
            elif message.metadata.data_encoding == 'gzip_proto':
                obj.ParseFromString(zlib.decompress(message.data))
            elif message.metadata.data_encoding == 'base64_gzip_proto':  # deprecated
                obj.ParseFromString(zlib.decompress(base64.b64decode(message.data)))
            elif message.metadata.data_encoding == 'raw':
                raw_data = message.data

        if validator is not None:
            try:
                if hasattr(validator, '__func__'):
                    # for static functions
                    validator.__func__(obj)
                else:
                    # for non-static functions
                    validator(obj)
            except Exception as e:
                logging.getLogger('protocol.protobuf_base').warning(e.__str__())
                raise e

        return {
            'service': {
                'name': message.service.name,
                'uuid': message.service.uuid
            },
            'metadata': {
                'message_uuid': message.metadata.message_uuid,
                'protocol_version': message.metadata.protocol_version,
                'timestamp': message.metadata.timestamp,
                'topic': message.metadata.topic,
                'data_encoding': message.metadata.data_encoding
            },
            'data': raw_data if raw_data is not None else obj
        }, message_uuid, timestamp
